# viewbox.tcl
# Written by Winston Dang  wkd@hawaii.edu
# Jan 28, 1994
# Show viewer stats in a listbox
# Arguments:
#    w -	Name to use for new top-level window.

# The following is used for testing only 
#set username("winston") "winston"
#set good("winston") 3
#set bad("winston") 2
#set retries("winston") 1
#set packrecv("winston") 0
#set viewlock 1

# add a new user to viewer list box
proc memberadd {goodnum badnum retriesnum packnum user member} {
	global username good bad retries packrecv viewlock
	if {[info exists good($member)] == 0} {
		set username($member) $user
		set good($member) $goodnum
		set bad($member) $badnum
		set retries($member) $retriesnum
		set packrecv($member) $packnum
		set viewlock 1
	} else {
		set username($member) $user
		incr good($member) $goodnum
		incr bad($member) $badnum
		incr retries($member) $retriesnum
		incr packrecv($member) $packnum
		set viewlock 1
	}
}

# add a new server or increment it
proc serveradd {member servername namefile} {
	global server good bad retries packrecv filelog username
	if {[info exists good($member)] == 0} {
		set good($member) 0
		set bad($member) 0
		set retries($member) 0
		set packrecv($member) 0
		set username($member) $servername
	}	
	if {[info exists server($member)] == 0} {
		set server($member) 1
	} else {
		incr server($member)
	}
	if {[llength $filelog] > 48} {
		lreplace $filelog 0 0
	}
	lappend filelog "$namefile $servername"
	catch {.r1.log.list insert 0 "$namefile $servername"}
}

# clear all entries in listbox
proc memberreset {w} {
    global username good bad retries packrecv viewlock server	
	foreach user [array names good] {
		unset username($user)
		unset good($user)
		unset bad($user)
		unset retries($user)
		unset packrecv($user)
	}
	foreach user [array names server] {
		unset server($user)
	}
	set server() {}
	unset viewlock
	destroy $w
}

proc loginfo {w} {
	global good server
	set viewers [array size good]
      mkDialog .info "-text {Authored by Winston Dang} -aspect 400" \
               "OK {}"
}                     

proc mkViewbox2 {{w .v2}} {
    global username good bad retries packrecv viewlock server
    catch {destroy $w}
    toplevel $w
    wm title $w "Viewer Stats"
    wm iconname $w "Viewers"
    wm minsize $w 1 1
    frame $w.viewer -borderwidth 10
    label $w.viewer.label -anchor w -text [format\
	"Good Bad   Pack     Retr %20.20s %d %30.30s %d" \
 	"users" [array size good] "servers" [ expr [array size server] -1]]
    pack append $w.viewer $w.viewer.label {left top expand fillx} 
 		
    scrollbar $w.viewer.scroll -relief sunken -command "$w.viewer.list yview"
    listbox $w.viewer.list -yscroll "$w.viewer.scroll set" -relief sunken \
	    -setgrid 1
    pack append $w.viewer $w.viewer.scroll {right filly} \
	    $w.viewer.list {left expand fill}
    if {[info exists viewlock] == 1} {	
	foreach user [array names good] {
		if {[info exists server($user)] == 0} {	
			$w.viewer.list insert 0 \
			[format "%4.4d %3.3d %8.8d %4.4d %s" \
			$good($user) $bad($user) \
			$packrecv($user) $retries($user) $username($user)]
		} else {
			$w.viewer.list insert 0 \
			[format "%4.4d %3.3d %8.8d %4.4d %-40.40s ** %4.4d"\
			$good($user) $bad($user) \
			$packrecv($user) \
			$retries($user) $username($user) $server($user)]
		}			
	}
    }
    frame $w.butframe -borderwidth 10
    button $w.butframe.quit -text OK -command "destroy $w"
    pack append $w.butframe $w.butframe.quit {left expand fillx} \
	[button $w.butframe.info -command "loginfo $w"] \
	 {left} \
	[button $w.butframe.clear -text "clear table" \
	-command "memberreset $w"] {left expand fillx}  
    pack append $w $w.viewer {top expand filly frame center} \
	$w.butframe {bottom fillx}
}





