/* client program to perform local file operations */
/******************************************************************************
      IMM version 3.0
      Written by Winston Dang <wkd@hawaii.edu>
 
 * Copyright (c) University of Hawaii 1993. All rights reserved.
 *  
 * License is granted to copy, to use, and to make and to use derivative
 * works for research, educational, government or evaluation purposes, provided
 * that the University of Hawaii is acknowledged in all documentation pertaining
 * to any such copy or derivative work. University of Hawaii grants no other 
 * licenses expressed or implied. The University of Hawaii trade name should 
 * not be used in any advertising without its written  permission.
 *  
 * UNIVERSITY OF HAWAII MAKES NO REPRESENTATIONS CONCERNING EITHER THE
 * MERCHANTABILITY OF THIS SOFTWARE OR THE SUITABILITY OF THIS SOFTWARE
 * FOR ANY PARTICULAR PURPOSE.  The software is provided "as is" without
 * express or implied warranty of any kind.
 *  
 * These notices must be retained in any copies of any part of this software.
 ******************************************************************************/
/*
 * Portions of this code were written by Joe Macker,  1996
 */
#include "protocol.h"
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#include <stdlib.h>
#include <unistd.h>

/* External references (that really should be in a /usr/include file) */
extern char *optarg;
extern int optind, opterr;
extern int useimageloader;
extern struct hosttype *hosts_tail;
#ifndef __NetBSD__
extern char *sys_errlist[];
#endif
extern unsigned int total;
extern int fullscrflag;
extern int xvflag;
extern int debug;
extern int checkpath();
extern int archfilenum;
#define TRANSMISSION_OK 1

struct listnode {
    char *name;
    struct listnode *next;
};
struct listnode ListFile[MAXFILES];
struct listnode *ListPtr;

/* External references to sockets */
extern int xmitsock;
extern int sock;
extern int serverid;

extern struct nodetype nodelist[MAXFILES];
extern struct stattype statnum;

char *immdirectory;		/* directory name to store files */
char *displ;
#define DEFAULT_IMAGE_DIRECTORY "/tmp"

/**********************************************************/
/* file initializations */                                          
int xlinit(int numfiles, char *displaywin)
{
    int i;
     /* check for IMMDIR environ var */
     immdirectory = getenv("IMM_IMAGE_DIR");
     if (immdirectory == NULL) {
	 immdirectory = DEFAULT_IMAGE_DIRECTORY;
     }
     if (debug) fprintf(logfp,"directory used: %s\n",immdirectory);

    /* build circular link list to hold file names */
    for (i=0; i < (MAXFILES - 1); i++) {
	ListFile[i].name = NULL;
	ListFile[i].next = &ListFile[i+1];
    }
    ListFile[MAXFILES - 1].name = NULL;
    ListFile[MAXFILES - 1].next = &ListFile[0];
    ListPtr = &ListFile[0];

    
         /* copy displaywin to displ var */
    displ = (char *)malloc( strlen(displaywin) + 1);
    strcpy(displ,displaywin);

    /* test to see what image loader we are using */
     if (useimageloader != NULL) {
	 if (checkpath(useimageloader)) return;
     }
     xvflag = checkpath("xv");
     if (xvflag) return;

     if (checkpath("xloadimage") ) return;            
     fatal("IMM Could not find xv or xloadimage\n");
 }

/**********************************************************/   
/* subroutine to open and close files returning handle to stdio */
int xlopen(redo, nodeptr)
int redo;
struct nodetype *nodeptr;
{
    static unsigned int filecounter = 0;
    switch (redo) {
    case 1:
	close(nodeptr->ofp);
	return;
    default:
 	sprintf(nodeptr->tmpfile,"%s/.in%d.imm",
		     immdirectory, filecounter++);
	nodeptr->ofp = creat(nodeptr->tmpfile,0640);
	if (nodeptr->ofp != -1) 
	    return(nodeptr->ofp);
	printf("Unable to Open %s\n",nodeptr->tmpfile);
	exit(0);
    }
}
/**********************************************************/
/* change the user name */
int chghostname(mynewhostname)
char *mynewhostname;
{
    if (mynewhostname == NULL) 
	statnum.name_user[0] = NULL;
    else
	strncpy(statnum.name_user,mynewhostname,(MAXUSERNAME -1));
}
/**********************************************************/   
/* create subdirectory if they don't exist */
int makesubdir(direct)
char *direct;
{
    char *token;
    char lasttoken[256];
    char *workdir;
    char tokensep[] = "//";	/* token seperator */
    int retval = TRUE;
    workdir = getenv("PWD");
    if (chdir(immdirectory)) {
	warn("failed to chdir to %s\n",immdirectory);
	return FALSE;
	}
    token = strtok(direct,tokensep);
    strcpy(lasttoken,token);
    token = strtok(NULL,tokensep);
    
    while (token != NULL) {
      if ( chdir(lasttoken) ) {
          if (mkdir(lasttoken,0755)) {
              warn("unable to create %s\n",direct);
              retval = FALSE;
	      break;
          }
          chdir(lasttoken);
          if (debug > 2) fprintf(logfp,"create subdir %s\n",lasttoken);
      }
      strcpy(lasttoken,token);      
      token = strtok(NULL,tokensep);
    }
    if (chdir(workdir))
	warn("failed to chdir to %s\n",workdir);
    return retval;
}

/**********************************************************/   
/* close file and rename it, put it in the proper subdirect */
int xlclose(nodeptr)
struct nodetype *nodeptr;
{
    char cmdline[250];
    char xvopts[200];
    int i;
    if (nodeptr->transmitcode == TRANSMISSION_REST) return FALSE;
    nodeptr->transmitcode = TRANSMISSION_REST;
    close(nodeptr->ofp);
    
    statnum.lastreceived = nodeptr;
    sprintf(cmdline,"%s/%s",
	    immdirectory, nodeptr->currfile);
    if (debug) printf("rename %s->%s\n",nodeptr->tmpfile,cmdline);
    if ( ! access(cmdline,F_OK)) {
	/* file exists add extension counter */
	for (i=1; i < MAXFILES; i++) {
	    sprintf(cmdline,"%s/%s.%d",
		    immdirectory, nodeptr->currfile,i);
	    if (debug) printf("duplicate %s->%s\n",nodeptr->tmpfile,cmdline);
	    if ( access(cmdline,F_OK)) break;
	}
	strcpy(xvopts,nodeptr->currfile);
	sprintf(nodeptr->currfile,"%s.%d",xvopts, i);
    }
    if (rename(nodeptr->tmpfile,cmdline)) {
	if (makesubdir(nodeptr->currfile)) {
	    if (rename(nodeptr->tmpfile,cmdline))
		fprintf(logfp, "unable to change %s to %s\n",
			nodeptr->tmpfile,cmdline);
	}
    }
    if (nodeptr->serverflags & BUF_FLG_EXECUT) {
	chmod(cmdline, S_IXUSR | S_IXGRP | S_IRUSR | S_IRGRP);
    }
    nodeptr->tmpfile[0] = NULL;

    /* file received - now do some post processing with it */
    fullscrflag = build_xvopts(xvflag, xvopts);

    if (fullscrflag != 11) {
	ListPtr = ListPtr->next;
	if (ListPtr->name != NULL) {
	    if (debug) fprintf(logfp,"unlink %s\n",ListPtr->name);
	    if (archfilenum != 0) unlink(ListPtr->name);
	    free(ListPtr->name);
	}
	ListPtr->name  = malloc(strlen(cmdline) + 1);
	strcpy(ListPtr->name,cmdline);
    }
    

    switch (fullscrflag) {
    case TRUE:
      if (xvflag) {
          sprintf(cmdline,
                  "xv %s -root -disp %s -quit %s/%s",
                  xvopts,displ,immdirectory,nodeptr->currfile);
	  post_process(cmdline);
      } else {
          sprintf(cmdline,
                  "xloadimage %s  -onr -displ %s -bor %s %s/%s",
                  xvopts,displ,immdirectory,nodeptr->currfile);
	  post_process(cmdline);
      }
      break;
    case FALSE:
          if (xvflag) {
              sprintf(cmdline,
                      "xv -rmod 0 -root -disp %s -maxp -quit %s",
                      displ,ListPtr->name);
	      post_process(cmdline);
          } else {
              sprintf(cmdline,
                      "xloadimage -onr -displ %s -bor black -full %s",
                      displ,ListPtr->name);
	      post_process(cmdline);
	  }
      break;
    case 10:
      sprintf(cmdline,
              "%s -disp %s -quit %s",
              xvopts,displ,ListPtr->name);
      post_process(cmdline);
      break;
    case 11:
      return TRUE;
  case 12:
      sprintf(cmdline,
              "%s %s",
              xvopts,ListPtr->name);
      post_process(cmdline);
      break;
    }
    if (debug) fprintf(logfp,"%s\n",cmdline); 
    return TRUE;
}
/**********************************************************/
/* signal interrupt to abort image - close up shop */
int xlredraw()
{
    char cmdline[250];
    char xvopts[40];

    if (ListPtr->name == NULL) return FALSE;
    fullscrflag = build_xvopts(xvflag, xvopts);
    switch (fullscrflag) {
    case 10:
	sprintf(cmdline,
		"%s -disp %s %s",
		xvopts,displ,ListPtr->name);
	post_process(cmdline);
	break;
    case 12:
	sprintf(cmdline,
		"%s %s",
		xvopts,ListPtr->name);
	post_process(cmdline);
	break;
	
    default:
	if (xvflag) {
	    sprintf(cmdline,
		    "xv -disp %s %s",
		    displ,ListPtr->name);
	    post_process(cmdline);
	} else {
	    sprintf(cmdline,
		    "xloadimage -displ %s %s",
		    displ,ListPtr->name);
	    post_process(cmdline);
	}
	break;
    }
    if (debug) fprintf(logfp,"%s\n",cmdline); 
    return TRUE;
}
/**********************************************************/
int post_process(char *cmdline) {
    int pid = 0;
/*    system(cmdline);
    return TRUE;
*/    
#ifdef SIGCLD	
	signal(SIGCLD,SIG_IGN);
#endif
#ifdef SIGCHLD
	signal(SIGCHLD,SIG_IGN);
#endif
    pid = fork();
    switch (pid) {
    case -1:
	fprintf(stderr, "unable to fork \n");
	exit(1);
    case 0:	/* Child process */

	checkhost(RECOV_MISS_FREE,NULL,NULL);
	fclose(stdin);
	close(xmitsock);
	close(sock);

	if (debug) fprintf( stderr,"post_process %s\n", cmdline);
	execlp("/bin/sh","/bin/sh","-c",cmdline,NULL);	

	exit(0);
    default:
	return TRUE;
    }
}    
/**********************************************************/           
/* signal interrupt to abort image - close up shop */
void abortimage()
{
    struct nodetype *nodeptr;
    struct hosttype *hostptr;
    int i;
    if (debug) printf("shutting down\n");
    fflush(stdout);
    if (fullscrflag != 11) {
	for (i=0; i < MAXFILES; i++) {
	    if (ListPtr->name != NULL) {
		if (debug) printf("unlink %s\n",ListPtr->name);
		unlink(ListPtr->name);
	    }
	    ListPtr = ListPtr->next;
	}
    }
    for (hostptr = hosts_tail; hostptr != NULL;
	hostptr = (struct hosttype *) hostptr->prev) {
	nodeptr = &hostptr->node;
	close(nodeptr->ofp);
	if (nodeptr->tmpfile[0] != NULL) unlink(nodeptr->tmpfile);
    }
    if (serverid) {
	fprintf(logfp,"kill server proc %d\n",serverid);
	kill(serverid,SIGKILL);
    }

    if (logfp)
	fclose(logfp);
    exit(0);
}

/* close all the files */ 
int closeall()
{
    int i;
    if (fullscrflag != 11) {    
	for (i=0; i < MAXFILES; i++) {
	    if (ListPtr->name != NULL) {
		unlink(ListPtr->name);
	    }
	    ListPtr = ListPtr->next;
	}
    }
    close(xmitsock);
    close(sock);
    fclose(logfp);
}







